#center {
    text-align: center;
}

#logo {
    width: 400px;
    height: auto;
    margin: auto;
    border: 0px;
    border-style: none;
    color: #ffffff00;
    background: #ffffff00;
    box-shadow: none;
}

.download {
    visibility: hidden;
}
