#center {
    text-align: center;
}

.download {
    visibility: hidden;
}

div#description-header{
    margin-left: auto;
    margin-right: auto;
    margin-bottom: .5em;
}

span#header-logo {
    width: 96px;
    height: 96px;
    display: inline-block;
    vertical-align: middle;
}

span#header-text {
    display: inline-block;
    vertical-align: middle;
}

div#id-score-table-container {
    text-align: center;
}

p#score-para {
    font-size: .8em;
}

span.span-minor {
    font-size: .8em;
}

span.span-major {
    font-weight: bold;
}

button.match-height{
    height: 100%;
}

div#rate-text {
    margin-top: 0.5em;
    margin-bottom: 2em;
    margin-left: auto;
    margin-right: auto;
    text-align: center;
    font-size: 1.5em;
}

div#description{
    max-width: 640px;
    margin: 0em auto 2em auto;
    display: grid;
}

.paper_row {
    margin-bottom: .8em;
}

.paper_link {
    font-size: 1em;
}

.paper_abstract {
    font-size: .85em;
}
